-- Put functions in this file to use them in several other scripts.
-- To get access to the functions, you need to put:
-- require "my_directory.my_file"
-- in any script using the functions.

local map = {}

MAP_COLLISIONS_LAYER = "collisions"
MAP_ANIMATED_LAYER = "animated"
MAP_GROUND_LAYER = "ground"

MAP_COLLISION_BLOCKER_NAME = "blocker"
MAP_COLLISION_INTERACTIVE_NAME = "interactive"
MAP_COLLISION_NPC_NAME = "npc"
MAP_TRIGGER_TRANSPORT_NAME = "transport"
MAP_TRIGGER_ANIMATION_NAME = "animated"

MAP_TILE_GRASS_0 = 7449

TILE_SIZE = 16

local map_width = nil
local map_height = nil
local map_rows = nil
local map_columns = nil
local map_tiles = nil

local current_map = nil

local global_submaps = {}

--function map.init(self)
--	global_submaps[self.map_id] = self.sub_maps
--end

function map.submap_init(self)
	if global_submaps[self.parent_tilemap_id] == nil then
		global_submaps[self.parent_tilemap_id] = {}
	end
	global_submaps[self.parent_tilemap_id][self.tilemap_id] = {
		c0 = self.c0,
		r0 = self.r0,
		c1 = self.c1,
		r1 = self.r1,
		label = self.label
	}
end

function map.get_submaps(mapurl)
	return global_submaps[mapurl]
end

function map.get_submap_from_tileid(mapurl, maptileid)

	if (mapurl == nil) then
		mapurl = map.get_current()
	end

	r = map.get_tile_row(mapurl, maptileid)
	c = map.get_tile_column(mapurl, maptileid)

	if (map.get_submap() == nil or not
	(map.get_submap_c0() <= c and map.get_submap_c1() >= c and
	map.get_submap_r0() <= r and map.get_submap_r1() >= r)) then

		if (global_submaps[mapurl] ~= nil) then
		
			for k, v in pairs(global_submaps[mapurl]) do

				if (v['c0'] <= c and v['c1'] >= c and
				v['r0'] <= r and v['r1'] >= r) then

					if (map.get_submap() ~= k ) then
						return k
						--map.set_submap(k,  v['c0'],  v['c1'],  v['r0'],  v['r1'],  v['label'])
						--msg.post("/gui#gamepad", "current_map", { label = self.label .. " - " .. v['label'] })
						--print(v['label'])
					end

				else

					if (map.get_submap() ~= nil ) then

						return nil
						--map.set_submap(nil, nil, nil, nil, nil, "Void")
						--msg.post("/gui#gamepad", "current_map", { label = "Void" })
						--print(map.get_submap_label())

					end

				end
			end

		else

			if (map.get_submap() ~= nil ) then

				return nil
			end

		end

	elseif (map.get_submap_c0() <= c and map.get_submap_c1() >= c and
	map.get_submap_r0() <= r and map.get_submap_r1() >= r) then
		return map.get_submap()
	end

end

function map.hide_map(url)
	if( (url == current_map or url == nil) and current_map ~= nil ) then
		p = go.get_position(current_map)
		p.z = 100
		go.set_position(p, current_map)
	elseif ( url ~= nil ) then
		p = go.get_position(url)
		p.z = 100
		go.set_position(p, url)
	else
		print("ERROR: no map selected")
	end
end

function map.show_map(url)
	if( (url == current_map or url == nil) and current_map ~= nil ) then
		p = go.get_position(current_map)
		p.z = 0
		go.set_position(p, current_map)
	elseif ( url ~= nil ) then
		p = go.get_position(url)
		p.z = 0
		go.set_position(p, url)
	else
		print("ERROR: no map selected")
	end
end

function map.set_current(url)
	
	if (current_map ~= nil) then
		map.hide_map(current_map)
	end
	
	current_map = url
	map.show_map(current_map)

	map_width = nil
	map_height = nil
	map_rows = nil
	map_columns = nil
	map_tiles = nil
	
end

function map.get_current()
	return current_map
end

function map.get_x(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then
		return go.get_position(current_map).x
	elseif ( url ~= nil ) then
		return go.get_position(url).x
	else
		print("ERROR: no map selected")
		return nil
	end
	
end

function map.get_y(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then
		return go.get_position(current_map).y
	elseif ( url ~= nil ) then
		return go.get_position(url).y
	else
		print("ERROR: no map selected")
		return nil
	end

end

function map.get_columns(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then

		if(map_columns == nil) then
			local x, y, c, r = tilemap.get_bounds(current_map)
			map_columns = c
		end
		
		return map_columns

	elseif ( url ~= nil ) then
		local x, y, c, r = tilemap.get_bounds(url)
		return c
	else
		print("ERROR: no map selected")
		return nil
	end

end

function map.get_rows(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then

		if(map_rows == nil) then
			local x, y, c, r = tilemap.get_bounds(current_map)
			map_rows = r
		end

		return map_rows

	elseif ( url ~= nil ) then
		local x, y, c, r = tilemap.get_bounds(url)
		return r
	else
		print("ERROR: no map selected")
		return nil
	end

end

function map.get_tiles(url)
	if(map.get_rows(url) == nil or map.get_columns(url) == nil) then
		return nil
	else
		return map.get_rows(url) * map.get_columns(url)
	end
end

function map.get_width(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then

		if(map_width == nil) then
			local x, y, c, r = tilemap.get_bounds(current_map)
			map_columns = c
			map_width = c * TILE_SIZE
		end

		return map_width
		
	elseif ( url ~= nil ) then
		local x, y, c, r = tilemap.get_bounds(url)
		return c * TILE_SIZE
	else
		print("ERROR: no map selected")
		return nil
	end
	
end

function map.get_height(url)

	if( (url == current_map or url == nil) and current_map ~= nil ) then

		if(map_height == nil) then
			local x, y, c, r = tilemap.get_bounds(current_map)
			map_rows = r
			map_height = r * TILE_SIZE
		end

		return map_height

	elseif ( url ~= nil ) then
		local x, y, c, r = tilemap.get_bounds(url)
		return r * TILE_SIZE
	else
		print("ERROR: no map selected")
		return nil
	end

end

function map.get_tile_size()

	return TILE_SIZE

end

function map.get_tile_x(url, id)

	if(id == nil) then
		print("ERROR: no tile id selected")
		return nil
	end

	if(id < 0) then
		print("ERROR: this map as minimum tile id = 0")
		return nil
	end

	if(id >= map.get_tiles(url) ) then
		print("ERROR: this map as maximum tile id = " .. map.get_tiles(url) - 1)
		return nil
	end
	
	return (id % map.get_columns(url)) * map.get_tile_size()
	
end

function map.get_tile_y(url, id)

	if(id == nil) then
		print("ERROR: no tile id selected")
		return nil
	end

	if(id < 0) then
		print("ERROR: this map as minimum tile id = 0")
		return nil
	end

	if(id >= map.get_tiles(url) ) then
		print("ERROR: this map as maximum tile id = " .. map.get_tiles(url) - 1)
		return nil
	end
	
	return math.floor(id / map.get_columns(url)) * map.get_tile_size()
	
end

function map.set_position(url, tileid)
	
	if( (url == current_map or url == nil) and current_map ~= nil ) then

		go.set_position(vmath.vector3(-map.get_tile_x(nil, tileid), map.get_tile_y(nil, tileid), 0), current_map)

	elseif ( url ~= nil ) then

		go.set_position(vmath.vector3(-map.get_tile_x(url, tileid), map.get_tile_y(url, tileid), 0), url)
		
	else
		print("ERROR: no map selected")
		return nil
	end
	
end

function map.get_position(url, tileid)
	
	if( (url == current_map or url == nil) and current_map ~= nil ) then

		if(tileid == nil) then
			return go.get_position(current_map)
		end
		
		return vmath.vector3(-map.get_tile_x(nil, tileid), map.get_tile_y(nil, tileid), 0)

	elseif ( url ~= nil ) then

		if(tileid == nil) then
			return go.get_position(url)
		end
		
		return vmath.vector3(-map.get_tile_x(url, tileid), map.get_tile_y(url, tileid), 0)

	else
		print("ERROR: no map selected")
		return nil
	end
	
end

function map.get_tilemap_tileid(url, layer, tileid)
	
	if(tileid == nil) then
		print("ERROR: no tile selected")
		return nil
	end
	
	if( url == nil and current_map ~= nil ) then
		url = current_map
	elseif (url == nil and current_map == nil) then
		print("ERROR: no map selected")
		return nil
	end
	
	x = (map.get_tile_x(url, tileid)/16) + 1
	y = -(map.get_tile_y(url, tileid)/16)

	return tilemap.get_tile(url, layer, x, y)
end

function map.has_collision_group(tilemap_tileid, collision_group)

	return false
end

--function map.is_animated(mapurl, submapurl, maptileid)

--	if(submapurl == nil and map.get_submap() ~= nil) then
--		submapurl = map.get_submap()
--	elseif(submapurl == nil and map.get_submap() == nil) then
--		print("ERROR: no submap selected")
--		return nil
--	end
	
--	tilemap_tileid = map.get_tilemap_tileid(url, MAP_ANIMATED_LAYER, tileid)

--	return map.has_collision_group(tilemap_tileid, MAP_TRIGGER_ANIMATION_NAME)
--end

function map.is_transport(mapurl, submapurl, maptileid)

	if(submapurl == nil and map.get_submap() ~= nil) then
		submapurl = map.get_submap()
	elseif(submapurl == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	submap_tileid = map.get_submap_tileid(mapurl, submapurl, maptileid)
	
	tilemap_tileid = map.get_tilemap_tileid(submapurl, MAP_COLLISIONS_LAYER, submap_tileid)

	return map.has_collision_group(tilemap_tileid, MAP_TRIGGER_TRANSPORT_NAME)
end

function map.is_interactive_walkable(tilemap_tileid)
	
	if (tilemap_tileid == MAP_TILE_GRASS_0) then
		return true
	end
	
	return false
end

--function map.get_submap_from_tileid(mapurl, maptileid)
--	if(mapurl == nil) then
--		mapurl = map.get_current()
--	end
--	return map_common.get_submap(mapurl, maptileid)
--end

function map.is_walkable(mapurl, submapurl, maptileid)

	if(submapurl == nil and map.get_submap() ~= nil) then
		submapurl = map.get_submap()
	elseif(submapurl == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	submap_tileid = map.get_submap_tileid(mapurl, submapurl, maptileid)
	
	tilemap_tileid = map.get_tilemap_tileid(submapurl, MAP_COLLISIONS_LAYER, submap_tileid)
	
	if (map.has_collision_group(tilemap_tileid, MAP_COLLISION_BLOCKER_NAME)) then
		return false
	end

	if (map.has_collision_group(tilemap_tileid, MAP_COLLISION_NPC_NAME)) then
		return false
	end

	if(map.has_collision_group(tilemap_tileid, MAP_COLLISION_INTERACTIVE_NAME)) then
		--tilemap_tileid = map.get_tilemap_tileid(url, MAP_GROUND_LAYER, tileid)
		--return map.is_interactive_walkable(tilemap_tileid)
		return true
	end

	return true
end

function map.is_walkable_submap(submapurl, tileid)
	if(submapurl == nil and map.get_submap() ~= nil) then
		submapurl = map.get_submap()
	elseif(submapurl == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	tilemap_tileid = map.get_tilemap_tileid(submapurl, MAP_COLLISIONS_LAYER, tileid)

	if (map.has_collision_group(tilemap_tileid, MAP_COLLISION_BLOCKER_NAME)) then
		return false
	end

	if (map.has_collision_group(tilemap_tileid, MAP_COLLISION_NPC_NAME)) then
		return false
	end

	if(map.has_collision_group(tilemap_tileid, MAP_COLLISION_INTERACTIVE_NAME)) then
		--tilemap_tileid = map.get_tilemap_tileid(url, MAP_GROUND_LAYER, tileid)
		--return map.is_interactive_walkable(tilemap_tileid)
		return true
	end

	return true
	
end

function map.get_tileid_from_editor_position(url, x, y)
	if(map.get_columns(url) == nil) then
		print("ERROR: no map selected")
		return nil
	end
	return (x - 1) + (y * map.get_columns(url))
end

function map.get_tile_row(url, tileid)
	p = map.get_tile_y(url, tileid)
	if (p == nil) then
		return nil
	end
	return p/TILE_SIZE
end

function map.get_tile_column(url, tileid)
	p = map.get_tile_x(url, tileid)
	if (p == nil) then
		return nil
	end
	return (p/TILE_SIZE) + 1
end

local current_sub_map = nil
local current_sub_map_c0 = nil
local current_sub_map_r0 = nil
local current_sub_map_c1 = nil
local current_sub_map_r1 = nil
local current_sub_map_label = nil

function map.set_submap(url)
	current_sub_map = url
end

function map.get_submap()
	return current_sub_map
end

function map.get_submap_c0(mapurl, submap)
	
	if (mapurl == nil and map.get_current() ~= nil) then
		mapurl = map.get_current()
	elseif(mapurl == nil and map.get_current() == nil) then
		print("ERROR: no map selected")
		return nil
	end

	if (submap == nil and map.get_submap() ~= nil) then
		submap = map.get_submap()
	elseif(submap == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end
	
	return global_submaps[mapurl][submap]["c0"]
end

function map.get_submap_c1(mapurl, submap)
	
	if (mapurl == nil and map.get_current() ~= nil) then
		mapurl = map.get_current()
	elseif(mapurl == nil and map.get_current() == nil) then
		print("ERROR: no map selected")
		return nil
	end

	if (submap == nil and map.get_submap() ~= nil) then
		submap = map.get_submap()
	elseif(submap == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	return global_submaps[mapurl][submap]["c1"]
end

function map.get_submap_r0(mapurl, submap)
	
	if (mapurl == nil and map.get_current() ~= nil) then
		mapurl = map.get_current()
	elseif(mapurl == nil and map.get_current() == nil) then
		print("ERROR: no map selected")
		return nil
	end

	if (submap == nil and map.get_submap() ~= nil) then
		submap = map.get_submap()
	elseif(submap == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	return global_submaps[mapurl][submap]["r0"]
end

function map.get_submap_r1(mapurl, submap)
	
	if (mapurl == nil and map.get_current() ~= nil) then
		mapurl = map.get_current()
	elseif(mapurl == nil and map.get_current() == nil) then
		print("ERROR: no map selected")
		return nil
	end

	if (submap == nil and map.get_submap() ~= nil) then
		submap = map.get_submap()
	elseif(submap == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	return global_submaps[mapurl][submap]["r1"]
end

function map.get_submap_label(mapurl, submap)
	
	if (mapurl == nil and map.get_current() ~= nil) then
		mapurl = map.get_current()
	elseif(mapurl == nil and map.get_current() == nil) then
		print("ERROR: no map selected")
		return nil
	end

	if (submap == nil and map.get_submap() ~= nil) then
		submap = map.get_submap()
	elseif(submap == nil and map.get_submap() == nil) then
		print("ERROR: no submap selected")
		return nil
	end

	return global_submaps[mapurl][submap]["label"]
end

function map.get_submap_tileid(mapurl, submapurl, maptileid)
	
	if (mapurl == nil and current_map ~= nil) then
		mapurl = current_map
	elseif(mapurl == nil and current_map == nil) then
		print("ERROR: no map selected")
		return nil
	end
	
	if(submapurl == nil and current_sub_map ~= nil) then
		submapurl = current_sub_map
	elseif(submapurl == nil and current_sub_map == nil) then
		print("ERROR: no submap selected")
		return nil
	end
	
	c = map.get_tile_column(mapurl, maptileid)
	r = map.get_tile_row(mapurl, maptileid)
	c0 = map.get_submaps(mapurl)[submapurl]['c0']
	r0 = map.get_submaps(mapurl)[submapurl]['r0']
	return  map.get_tileid_from_editor_position(submapurl, (c - c0) + 1, r - r0)
end

function map.get_tileid_from_position(url, x, y)
	c = (x / TILE_SIZE) + 1
	r = -(y / TILE_SIZE)
	print(x .. "; " .. r)
	return map.get_tileid_from_editor_position(url, c, r)
end

return map